/*
	KLibrary : kthread.c

	kernel thread library

	collected and modified by flyduck 2001/05/18
	
	original sources are from http://www.scs.ch/~frey/linux/
	Original writer is Martin Frey, frey@scs.ch
*/

#define TESTLIBRARY				0

#if TESTLIBRARY == 0
#define __NO_VERSION__
#endif

#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/smp_lock.h>

#include <kthread.h>

//
// Function prototypes
//

void klib_start_kthread(kthread_func_t func, kthread_t *kthread, void *param);
void klib_stop_kthread(kthread_t *kthread);
void klib_init_kthread(kthread_t *kthread, char *name);
void klib_exit_kthread(kthread_t *kthread);

static void klib_kthread_launcher(void *data);

// 
// Library Functions
//

// klib_start_kthread : 
//   create a new kernel thread. Called by the creator.
// Context : In 2.4 : any, In 2.2 : not sure
// Parameter : 
//   func : (in) function to be executed as thread
//   kthread : (in) kthread data structure
//   param : (in) parameter to be passed to kthread data structure
// Return : none

void klib_start_kthread(kthread_func_t func, kthread_t * kthread, void *param)
{
    /* initialize the semaphore:
       we start with the semaphore locked. The new kernel
       thread will setup its stuff and unlock it. This
       control flow (the one that creates the thread) blocks
       in the down operation below until the thread has reached
       the up() operation.
     */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)	
    init_MUTEX_LOCKED(&kthread->startstop_sem);
#else
	kthread->startstop_sem = MUTEX_LOCKED;
#endif	

    /* store the function to be executed in the data passed to
       the launcher */
    kthread->function = func;

	/* store the parameter to be passed to handler */
	kthread->param = param;

    /* create the new thread my running a task through keventd */

    /* initialize the task queue structure */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)	
    kthread->tq.sync = 0;
    INIT_LIST_HEAD(&kthread->tq.list);
    kthread->tq.routine = klib_kthread_launcher;
    kthread->tq.data = kthread;

    /* and schedule it for execution */
    schedule_task(&kthread->tq);
#else
	klib_kthread_launcher(kthread);
#endif	

    /* wait till it has reached the setup_thread routine */
    down(&kthread->startstop_sem);
}

// klib_stop_kthread :
//   stop a kernel thread. Called by the removing instance 
// Context : Any
// Parameter : 
//   kthread : (in) kthread data structure
// Return : none

void klib_stop_kthread(kthread_t * kthread)
{
    if (kthread->thread == NULL) {
		printk("stop_kthread: killing non existing thread!\n");
		return;
    }

    /* this function needs to be protected with the big
       kernel lock (lock_kernel()). The lock must be
       grabbed before changing the terminate
       flag and released after the down() call. */
    lock_kernel();

    /* initialize the semaphore. We lock it here, the
       leave_thread call of the thread to be terminated
       will unlock it. As soon as we see the semaphore
       unlocked, we know that the thread has exited.
     */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)	
    init_MUTEX_LOCKED(&kthread->startstop_sem);
#else
	kthread->startstop_sem = MUTEX_LOCKED;
#endif	

    /* We need to do a memory barrier here to be sure that
       the flags are visible on all CPUs. 
     */
    mb();

    /* set flag to request thread termination */
    kthread->terminate = 1;

    /* We need to do a memory barrier here to be sure that
       the flags are visible on all CPUs. 
     */
    mb();
    kill_proc(kthread->thread->pid, SIGKILL, 1);

    /* block till thread terminated */
    down(&kthread->startstop_sem);

    /* release the big kernel lock */
    unlock_kernel();

    /* now we are sure the thread is in zombie state. We
       notify keventd to clean the process up.
     */
    kill_proc(2, SIGCHLD, 1);
}

// klib_init_kthread : 
//   initialize new created thread. Called by the new thread. 
// Context : Inside the kernel thread. This function should be called first 
// Parameter : 
//   kthread : (in) kthread data structure
//   name : (in) name of process
// Return : none

void klib_init_kthread(kthread_t * kthread, char *name)
{
    /* lock the kernel. A new kernel thread starts without
       the big kernel lock, regardless of the lock state
       of the creator (the lock level is *not* inheritated)
     */
    lock_kernel();

    /* fill in thread structure */
    kthread->thread = current;

    /* set signal mask to what we want to respond */
    siginitsetinv(&current->blocked,
		  sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGTERM));

    /* initialise wait queue */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)	
    init_waitqueue_head(&kthread->queue);
#else
	kthread->queue = NULL;
#endif	

    /* initialise termination flag */
    kthread->terminate = 0;

    /* set name of this process (max 15 chars + 0 !) */
    sprintf(current->comm, name);

    /* let others run */
    unlock_kernel();

    /* tell the creator that we are ready and let him continue */
    up(&kthread->startstop_sem);
}

// klib_exit_kthread : 
//   cleanup of thread. Called by the exiting thread. 
// Context : Inside the kernel thread. This function should be called right before exiting
// Parameter : 
//   kthread : (in) kthread data structure
// Return : none

void klib_exit_kthread(kthread_t * kthread)
{
    /* we are terminating */

    /* lock the kernel, the exit will unlock it */
    lock_kernel();
    kthread->thread = NULL;
    mb();

    /* notify the stop_kthread() routine that we are terminating. */
    up(&kthread->startstop_sem);

    /* the kernel_thread that called clone() does a do_exit here. */

    /* there is no race here between execution of the "killer" and real termination
       of the thread (race window between up and do_exit), since both the
       thread and the "killer" function are running with the kernel lock held.
       The kernel lock will be freed after the thread exited, so the code
       is really not executed anymore as soon as the unload functions gets
       the kernel lock back.
       The init process may not have made the cleanup of the process here,
       but the cleanup can be done safely with the module unloaded.
     */
}

// klib_kthread_launcher
//   Real function launching kernel thread.
//   This function is called by task queue handler (in 2.4)

void klib_kthread_launcher(void *data)
{
    kthread_t *kthread = data;

    kernel_thread((int (*)(void *)) kthread->function, (void *) kthread, 0);
}

//
// Test Functions
// 

#if TESTLIBRARY

static kthread_t kthread;

static int example_thread(kthread_t *pthread);

int init_module(void)
{
	klib_start_kthread(example_thread, &kthread, NULL);
	
	return 0;
}

void cleanup_module(void)
{
	klib_stop_kthread(&kthread);
}

int example_thread(kthread_t *pthread) 
{
	int loop;
	
	printk("KLIB_KTHREAD : starting thread loop\n");
	
	klib_init_kthread(&kthread, "Example thread");
	
	for (loop = 0;; ++loop) {
		printk("Thread Loop %d\n", loop);

		interruptible_sleep_on_timeout(&pthread->queue, HZ);
		
		if (pthread->terminate)
			break;
	}
	
	printk("KLIB_THREAD : exiting thread loop\n");
	
	klib_exit_kthread(&kthread);
	
	return 0;
}

#endif

